package scalax.collection
package immutable

import collection.Set
import collection.mutable.{Set => MutableSet, SetLike => MutableSetLike}

import scalax.collection.{Graph => CommonGraph}
import GraphEdge.{EdgeLike, EdgeCompanionBase}
import GraphPredef.EdgeLikeIn
import config.{GraphConfig, AdjacencyListArrayConfig}
import mutable.ArraySet
import generic.{GraphCompanion, GroupIterator, AdjacencyListBase}
import io._

/**
 * Implements an adjacency list based graph representation.
 * 
 * An adjacency list based representation speeds up traversing a graph along its paths
 * by storing the list of connecting edges to each node.
 *   
 * @author Peter Empen
 */
trait AdjacencyListGraph[N,
                         E[X] <: EdgeLikeIn[X],
                        +This[X, Y[X]<:EdgeLikeIn[X]]
                              <: AdjacencyListGraph[X,Y,This] with Graph[X,Y]]
  extends GraphLike[N,E,This]
  with    GraphAux [N,E]
  with    AdjacencyListBase[N,E,This]
{ selfGraph: This[N,E] =>
  protected type Config <: GraphConfig with AdjacencyListArrayConfig

  trait InnerNodeImpl extends InnerNodeLike {
    this: NodeT =>
    @inline final override def edges: Set[EdgeT] = nodes edges this
  }
  type NodeSetT = NodeSet
  @SerialVersionUID(7870L)
  class NodeSet extends super.NodeSet with NodeSetAux
  {
    @transient implicit val hints: ArraySet.Hints = config.adjacencyListHints
    @SerialVersionUID(7871L)
    protected class MapExt extends collection.mutable.HashMap[NodeT, ArraySet[EdgeT]]{
      @inline final override def findEntry(key: NodeT) = super.findEntry(key)
    }
    protected val coll = new MapExt
    @inline protected final def emptyEdges = ArraySet.ungrowableEmpty[EdgeT]
    override def copy = {
      val nodeSet = new NodeSet 
      nodeSet initialize (this.toNodeInSet, null)
      nodeSet
    }
    override protected[collection] def initialize(
        nodes: collection.Iterable[N],
        edges: collection.Iterable[E[N]]) = {
      if (nodes ne null) 
        coll ++= nodes map (n => (Node(n) -> emptyEdges))
      this
    }
    override protected[collection] def from (
        nodeStreams: Iterable[NodeInputStream[N]],
        nodes:       Iterable[N],
        edgeStreams: Iterable[GenEdgeInputStream[N,E]],
        edges:       Iterable[E[N]]) {
      for (n <- new NodeAux.NodeContStream(nodeStreams, nodes))
        coll += (Node(n) -> emptyEdges)
    }
    @inline final override def find(elem: N): Option[NodeT] = {
      val e = coll.findEntry(Node(elem))
      if (e == null) None else Some(e.key)
    }
    @inline final def edges(node: NodeT): Set[EdgeT] = {
      val e = coll.findEntry(node)
      if (e == null) Set.empty[EdgeT] else e.value
    }
    @inline final def contains(node: NodeT) =	coll.contains(node)
    @inline final def iterator: Iterator[NodeT] =	coll.keySet.iterator
    @inline final override protected def minus     (node: NodeT) { coll -= node }
    def +(node: NodeT) = { coll + (node -> emptyEdges); this }
    protected[collection] def += (edge: EdgeT): this.type = {
      for (n <- edge) {
        val entry = coll.findEntry(n)
        if (entry == null)                  coll += n  -> ArraySet(edge)
        else if (entry.value eq emptyEdges) entry.value = ArraySet(edge)
        else                                entry.value add edge
      }
      this
    }
  }
  protected val _nodes: NodeSet 
  @inline final def nodes = _nodes

  type EdgeT = EdgeBase
  @inline final override protected def newEdge(innerEdge: E[NodeT]) = new EdgeT(innerEdge)
  type EdgeSetT = EdgeSet
  @SerialVersionUID(7873L)
  class EdgeSet extends super.EdgeSet with EdgeSetAux
  {
    override protected[collection] def initialize(edges: collection.Iterable[E[N]]) {
      if (edges ne null)
        edges foreach (this += Edge(_))
    }
    override protected[collection]
    def from (nodeStreams: Iterable[NodeInputStream[N]],
              nodes:       Iterable[N],
              edgeStreams: Iterable[GenEdgeInputStream[N,E]],
              edges:       Iterable[E[N]])
    {
      for (e <- new EdgeAux.EdgeContStream(edgeStreams, edges))
        this += Edge(e)
    }
    final override def contains(node: NodeT): Boolean =
      _nodes find node exists (_.edges.nonEmpty)
    final override def find    (elem: E[N] ): Option[EdgeT] = {
      val e = Edge(elem)
      _nodes find (e._1) flatMap (_.edges find (_ == e))
    }
    @inline final protected[AdjacencyListGraph] def +=(edge: EdgeT): this.type = {
      _nodes += edge
      this
    }
    final def contains(edge: EdgeT): Boolean =
      _nodes find (edge._1) exists (_.edges contains edge)
    final def iterator: Iterator[EdgeT] = edgeIterator
    final override lazy val size = super.size
    @inline final def +(edge: EdgeT): Set[EdgeT] = toSet + edge
    @inline final def -(edge: EdgeT): Set[EdgeT] = toSet - edge
    @inline final override lazy val maxArity = super.maxArity
  }
  protected val _edges: EdgeSetT 
  @inline final def edges = _edges

  def copy(nodes: collection.Iterable[N],
           edges: collection.Iterable[E[N]]): This[N,E]
  def + (n: N)  = if (_nodes contains Node(n)) this
                  else copy(_nodes.toNodeInSet.toBuffer += n,
                            _edges.toEdgeInSet)
  protected
  def +#(e: E[N]) = if (_edges contains Edge(e)) this
                    else copy(_nodes.toNodeInSet,
                              _edges.toEdgeInSet.toBuffer += e)
  def - (n: N)  = _nodes find (nf => nf.value == n) match {
                    case Some(nf) => copy(_nodes.toNodeInSet.toBuffer  -= n,
                                          _edges.toEdgeInSet.toBuffer --= (nf.edges map (_.toEdgeIn)))
                    case None     => this
                  }
  def -?(n: N)  = _nodes find (nf => nf.value == n) match {
                    case Some(nf) => var newNodes = _nodes.toNodeInSet.toBuffer
                                     var newEdges = _edges.toEdgeInSet.toBuffer
                                     nodes.subtract(nf,
                                                    false,
                                                    nf => newNodes  -= n,
                                                    nf => newEdges --= (nf.edges map (_.toEdgeIn)))
                                     copy(newNodes, newEdges)
                    case None     => this
                  }
  protected
  def -#(e: E[N]) = if (_edges contains Edge(e))
                      copy(_nodes.toNodeInSet,
                           _edges.toEdgeInSet.toBuffer -= e)
                    else this
  protected
  def -!#(e: E[N]) = _edges find (ef => ef == e) match {
                    case Some(ef) => copy(_nodes.toNodeInSet.toBuffer --= ef.privateNodes map (n => n.value),
                                          _edges.toEdgeInSet.toBuffer  -= e)
                    case None     => this
                  }
}