package scalax.collection
package mutable

import collection.Set
import collection.mutable.{HashSet, HashMap, Set => MutableSet}

import scalax.collection.{Graph => CommonGraph}
import GraphEdge.{EdgeLike, EdgeCompanionBase}
import GraphPredef.EdgeLikeIn
import generic.{GraphCompanion, GroupIterator, AdjacencyListBase}
import config.{GraphConfig, AdjacencyListArrayConfig}
import io._

/**
 * Implements an adjacency list based graph representation.
 * 
 * An adjacency list based representation speeds up traversing a graph along its paths
 * by storing the list of connecting edges to each node.
 *   
 * @author Peter Empen
 */
trait AdjacencyListGraph[N,
                         E[X] <: EdgeLikeIn[X],
                        +This[X, Y[X]<:EdgeLikeIn[X]]
                              <: AdjacencyListGraph[X,Y,This] with Graph[X,Y]]
  extends GraphLike[N,E,This]
  with    AdjacencyListBase[N,E,This]
{
  protected type Config <: GraphConfig with AdjacencyListArrayConfig

  trait InnerNodeImpl extends InnerNodeLike {
    this: NodeT =>
    @inline final override def edges: Set[EdgeT] = nodes edges this
  }
  type NodeSetT = NodeSet
  @SerialVersionUID(7970L)
  class NodeSet extends super.NodeSet
  {
    @transient implicit val hints: ArraySet.Hints = config.adjacencyListHints
    @SerialVersionUID(7971L)
    protected class MapExt extends HashMap[NodeT, ArraySet[EdgeT]]{
      @inline final override def findEntry(key: NodeT) = super.findEntry(key)
    }
    protected val coll = new MapExt
    @inline protected final def emptyEdges = ArraySet.ungrowableEmpty[EdgeT]
    override def copy = {
      val nodeSet = new NodeSet
      nodeSet initialize (this.toNodeInSet, null)
      nodeSet
    }
    override protected[collection]
    def from (nodeStreams: Iterable[NodeInputStream[N]],
              nodes:       Iterable[N],
              edgeStreams: Iterable[GenEdgeInputStream[N,E]],
              edges:       Iterable[E[N]])
    {
      for (n <- new NodeAux.NodeContStream(nodeStreams, nodes))
        coll += (Node(n) -> emptyEdges)
    }
    override protected[collection]
    def initialize(nodes: Iterable[N],
                   edges: collection.Iterable[E[N]]) = {
      if (nodes ne null) 
        coll ++= nodes map (n => (Node(n) -> emptyEdges))
      this
    }
    @inline final override def find(elem: N): Option[NodeT] = {
      val e = coll.findEntry(Node(elem))
      if (e == null) None else Some(e.key)
    }
    @inline final def edges(node: NodeT): Set[EdgeT] = {
      val e = coll.findEntry(node)
      if (e == null) Set.empty[EdgeT] else e.value
    }
    @inline final def contains(node: NodeT) =	coll.contains(node)
    @inline final def iterator: Iterator[NodeT] =	coll.keySet.iterator
    override def add(node: NodeT) = if(coll.contains(node)) false
                                    else { coll += (node -> emptyEdges)
                                           true }
    @inline final def += (node: NodeT): this.type	= { add(node); this }
    protected[collection] def add (edge: EdgeT): Boolean = {
      // more performant than: "for (n <- edge) { this += n; coll(n) += edge }"
      var newEdge = true
      for (n <- Set(edge.toSeq: _*)) {
        val entry = coll.findEntry(n)
        if (entry == null)                  coll += n  -> ArraySet(edge)
        else if (entry.value eq emptyEdges) entry.value = ArraySet(edge)
        else {
          val added = entry.value add edge
          if (! added) newEdge = false
        }
      }
      newEdge
    }
    protected[collection] def += (edge: EdgeT): this.type	= { add(edge); this }
    protected[collection] def remove (edge: EdgeT): Boolean =
      edge forall (n => (coll get n) exists (_ remove edge))
    protected[collection] def -= (edge: EdgeT): this.type = { remove(edge); this }
    override protected def minus     (node: NodeT) { coll -= node }
    override protected def minusEdges(node: NodeT) { 
      // toList is necessary to avoid failure of -=(node) like in TEdit.test_MinusEq_2
      _edges --= node.edges.toList
    }
  }
  protected val _nodes: NodeSet 
  @inline final def nodes = _nodes

  
  type EdgeT = EdgeImpl
  @SerialVersionUID(7972L)
  class EdgeImpl(override val edge: E[NodeT]) extends EdgeBase(edge)
  {
    def remove: Boolean = _edges remove this
    def removeWithNodes(edge: E[N]) = if (_edges remove this) {
                                        _nodes --= privateNodes; true
                                      } else false
  }
  @inline final override protected def newEdge(innerEdge: E[NodeT]) = new EdgeT(innerEdge)
  type EdgeSetT = EdgeSet
  @SerialVersionUID(7974L)
  class EdgeSet extends super.EdgeSet
  {
    override protected[collection]def initialize(edges: Iterable[E[N]]) {
      if (edges ne null)
        edges foreach (this add Edge(_))
    }
    override protected[collection]
    def from (nodeStreams: Iterable[NodeInputStream[N]],
              nodes:       Iterable[N],
              edgeStreams: Iterable[GenEdgeInputStream[N,E]],
              edges:       Iterable[E[N]])
    {
      for (e <- new EdgeAux.EdgeContStream(edgeStreams, edges))
        this add Edge(e)
    }
    final override def contains(node: NodeT): Boolean =
      _nodes find node exists (_.edges.nonEmpty)
    final override def find    (elem: E[N] ): Option[EdgeT] = {
      val e = Edge(elem)
      _nodes find (e._1) flatMap (_.edges find (_ == e))
    }
    override def add(edge: EdgeT) =
      if (_nodes add edge) {
        nrEdges += 1
        true
      }
      else false
    final def contains(edge: EdgeT): Boolean =
      _nodes find (edge._1) exists (_.edges contains edge)
    final def iterator: Iterator[EdgeT] = edgeIterator
    override def remove(edge: EdgeT) = 
      if (_nodes remove edge) {
        nrEdges -= 1
        true
      } 
      else false
    def removeWithNodes(edge: EdgeT) = {
      val privateNodes = edge.privateNodes
      if (remove(edge)) {
        _nodes --= privateNodes
        true
      }
      else false
    }
    private var nrEdges = 0
    override def size = nrEdges

    @inline final override def maxArity: Int = super.maxArity
  }
  protected val _edges: EdgeSetT 
  @inline final def edges = _edges

  @inline final def clear	{	_nodes.clear }
  @inline final def add(node: N): Boolean = _nodes add Node(node)
  @inline final def add(edge: E[N]): Boolean  = _edges add Edge(edge)  
  @inline final protected def +=# (edge: E[N]): this.type = { add(edge); this }
}