package scalax.collection
package mutable

import collection.SortedSet
import collection.mutable.{Set => MutableSet, SetLike => MutableSetLike,
                           Builder, GrowingBuilder, ArrayBuffer}
import collection.generic.{CanBuildFrom, GenericSetTemplate, GenericCompanion,
                           MutableSetFactory}
import compat.Platform.arraycopy

import immutable.SortedArraySet

/** A basic [[ArraySet]] implementation suitable for efficient add operations.
 * Element removal could be optimized by another implementation.  
 * 
 * @param hints Optimization hints controlling the growth of the underlying
 *        [[scala.collection.mutable.Array]]. 
 */
@SerialVersionUID(1L)
class SimpleArraySet[A](override val hints: ArraySet.Hints)
  extends ArraySet[A]
  with    GenericSetTemplate[A, SimpleArraySet]
  with    MutableSetLike[A, SimpleArraySet[A]]
  with    Serializable
{
  override def companion: GenericCompanion[SimpleArraySet] = SimpleArraySet
  override def newBuilder = SimpleArraySet.newBuilderWithHints(this)
  override def clone = SimpleArraySet.emptyWithHints(hints.propagate(size)) ++=!
                       (if (isHash) hashSet else arr)
  private var nextFree: Int = 0
  private var arr: Array[A] = _
  private var hashSet: MutableSet[A] = _

  private def initialize {
    val capacity = hints.nextCapacity(0)
    if (capacity == 0) hashSet = MutableSet.empty[A]
    else               arr = new Array[AnyRef](capacity).asInstanceOf[Array[A]]
  }
  initialize
  final def capacity: Int = if (isHash) 0 else arr.length  
  @inline private def isHash: Boolean = arr eq null
  @inline final def isArray: Boolean = ! isHash
  protected[collection] def array = arr
  protected[collection] def set = hashSet

  def += (elem: A) = { add(elem); this }
  def -= (elem: A) = {
    if (isHash) hashSet -= elem
    else        removeIndex(indexOf(elem))
    this
  }
  protected def removeIndex(i: Int) {
    if (i != -1) {
      if (i + 1 < nextFree)
        arraycopy(arr, i + 1, arr, i, nextFree - i - 1)
      nextFree -= 1
    }
  }
  protected def +=!(elem: A): this.type = {
    if (isHash) hashSet add elem
    else {
      if (nextFree == capacity)
        if (resizedToHash) {
          add(elem); return this
        }
      arr(nextFree) = elem
      nextFree += 1
      true
    }
    this
  }
  protected def ++=!(xs: TraversableOnce[A]): this.type = {
    xs foreach +=! 
    this
  }
  override def iterator: Iterator[A] =
    if (isHash) hashSet.iterator
    else new Iterator[A] {
      private[this] var i = 0
      private[this] var prevElm: A = _
      def hasNext = {
        i < nextFree
      }
      def next = {
        prevElm = arr(i)
        i += 1
        prevElm
      }
    }
  override def foreach[U](f: (A) => U) {
    if (isHash) hashSet foreach f
    else {
      var i = 0
      while (i < nextFree) { f(arr(i)); i += 1 }
    }
  } 
  protected final def resizeArray(fromCapacity: Int, toCapacity: Int) {
    val newArr: Array[AnyRef] = new Array(toCapacity)
    arraycopy(arr, 0, newArr, 0, math.min(fromCapacity, toCapacity))
    arr = newArr.asInstanceOf[Array[A]]
  }
  protected final def setToArray(set: Iterable[A], size: Int) {
    arr = new Array[AnyRef](size).asInstanceOf[Array[A]]
    nextFree = 0
    set foreach { elem =>
      arr(nextFree) = elem
      nextFree += 1
    }
    hashSet = null
  }
  def compact {
    if (isHash) {
      val _size = size
      if (_size < hints.hashTableThreshold)
        setToArray(hashSet, _size)
    } else if (hints.compactUpToUsed match {
             case perc if perc ==   0 => false
             case perc if perc == 100 => nextFree < capacity
             case perc                => perc >= nextFree * 100 / capacity
           })
      resizeArray(capacity, nextFree)
  }
  protected final def indexOf(elem: A): Int = { // 'arr contains c' would be too slow
    var i = 0
    while (i < nextFree)
      if (arr(i) == elem) return i
      else i += 1
    -1
  }
  override def contains(elem: A): Boolean =
    if (isHash) hashSet contains elem
    else indexOf(elem) >= 0

  def find(elem: A): Option[A] =
    if (isHash) hashSet find (_ == elem)
    else {
      val i = indexOf(elem)
      if (i >= 0) Some(arr(i)) else None
    }
  override def add(elem: A): Boolean = {
    if (isHash) hashSet add elem
    else {
      if (nextFree == capacity)
        if (resizedToHash)
          return add(elem)
      var i = 0
      while (i < nextFree)
        if (arr(i) == elem) return false
        else i += 1
      arr(nextFree) = elem
      nextFree += 1
      true
    }
  }
  protected def resizedToHash: Boolean = {
    val newCapacity = hints.nextCapacity(capacity)
    if (newCapacity == 0) {
      hashSet = MutableSet.empty[A]
      hashSet sizeHint capacity
      hashSet ++= iterator
      arr = null
      true
    } else {
      resizeArray(capacity, newCapacity)
      false
    }
  }
  override def size = if (isHash) hashSet size
                      else nextFree
//  override def --=(toDelete: TraversableOnce[A]): this.type =
//    if (isHash) super.--=(toDelete)
//    else {
//      val deleteIndexes = new ArrayBuffer[Int](toDelete.size)
//      toDelete foreach { e =>
//        val i = indexOf(e)
//        if (i != -1) deleteIndexes += i - deleteIndexes.size
//      }
//      println("delete " + toDelete + " from " + this + " by " + deleteIndexes)
//      deleteIndexes foreach removeIndex
//      this
//    }
  // optimized for unchecked insertions
  override def filter(p: (A) => Boolean) = {
    val empty = SimpleArraySet.emptyWithHints(hints.propagate(size))
    if (isHash) super.filter(p)
    else {
      val b = SimpleArraySet.emptyWithPropagatedHints[A](this)
      for (x <- this)
        if (p(x)) b +=! x
      b
    }
  }
  def sorted(implicit ord: Ordering[A]): SortedSet[A] =
    if (isHash)
      hashSet match {
        case sorted: SortedSet[A] => sorted
        case unsorted             => SortedSet[A](hashSet.toList: _*)
      }
    else {
      val newArr: Array[AnyRef] = new Array(nextFree)
      arraycopy(arr, 0, newArr, 0, nextFree)
      new SortedArraySet(newArr.asInstanceOf[Array[A]])
    }
}
/**
 * @define FROM The [[ArraySet]] instance an operation of which this builder is invoked on.
 */
object SimpleArraySet extends MutableSetFactory[SimpleArraySet] {
  implicit def canBuildFrom[A]: CanBuildFrom[Coll, A, SimpleArraySet[A]] = setCanBuildFrom[A]
  /** Returns an empty set with default hints. */
  override def empty[A]: SimpleArraySet[A] = new SimpleArraySet[A](ArraySet.Hints())
  /** Returns an empty set with custom hints. */
  def emptyWithHints[A](implicit hints: ArraySet.Hints): SimpleArraySet[A] =
    new SimpleArraySet[A](hints)
  /** Returns an empty set with hints propagated from `arraySet`. */
  def emptyWithPropagatedHints[A](arraySet: ArraySet[A]): SimpleArraySet[A] =
    emptyWithHints(arraySet.hints.propagate(arraySet.size))
  /**
   * Returns a [[scala.collection.mutable.Builder]] implementation optimized for
   * building the new [[ArraySet]] based on the characteristics of `from`.
   * 
   * @param from $FROM 
   */
  protected
  def newBuilderWithHints[A] (from: ArraySet[A]): Builder[A, SimpleArraySet[A]] =
    new SimpleArraySetBuilder[A] (from)
  /**
   * Default `ArraySet` builder. The hints passed are propagated such that the `initial
   * size == to from.size`.
   * 
   * @param from $FROM 
   */
  protected class SimpleArraySetBuilder[A] (from: ArraySet[A])
    extends GrowingBuilder[A, SimpleArraySet[A]](emptyWithPropagatedHints(from))
}